import os
import sys
import snakemake

def snakemake_create_environments(args):

    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    snakefile_path = os.path.join(thisdir, "setup.smk")
    if os.path.exists(snakefile_path):
        snakefile = snakefile_path
        print(snakefile_path)
    else:
        msg = 'Error: cannot find Snakefile at the following location:\n'
        msg += '{}\n'.format(snakefile_path)
        sys.stderr.write(msg)
        sys.exit(-1)

    dependency_dir = os.path.join(args.dependency_dir, "ML_dependencies")
    workflow = None
    workflow = {"DEPENDENCYDIR": dependency_dir}
    
    status = snakemake.snakemake(snakefile,
                                config = workflow,
                                targets = ["all"],
                                conda_prefix = os.path.join(args.dependency_dir, "ML_dependencies", "ML_envs"),
                                use_conda = True,
                                conda_create_envs_only = True,
                                cores = 1, workdir = cwd)

     # Check if the workflow executed successfully
    if status:
        print("Conda environments were successfully created.")
    else:
        print("Creation of conda environments failed.")

def get_models(args):

    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    snakefile_path = os.path.join(thisdir, "setup.smk")
    if os.path.exists(snakefile_path):
        snakefile = snakefile_path
        print(snakefile_path)
    else:
        msg = 'Error: cannot find Snakefile at the following location:\n'
        msg += '{}\n'.format(snakefile_path)
        sys.stderr.write(msg)
        sys.exit(-1)

    dependency_dir = os.path.join(args.dependency_dir, "ML_dependencies")
    workflow = None
    workflow = {"DEPENDENCYDIR": dependency_dir}
    
    status = snakemake.snakemake(snakefile,
                                config = workflow,
                                targets = ["make_REbase_db"],
                                conda_prefix = os.path.join(args.dependency_dir, "ML_dependencies", "ML_envs"),
                                use_conda = True,
                                cores = 1, workdir = cwd)

     # Check if the workflow executed successfully
    if status:
        print("pfam models and REbase database were successfully retreived.")
    else:
        print("pfam models and REbase database could not be retreived.")


def defensefinder_update(args):

    thisdir = os.path.abspath(os.path.dirname(__file__))
    cwd = os.getcwd()

    snakefile_path = os.path.join(thisdir, "setup.smk")
    if os.path.exists(snakefile_path):
        snakefile = snakefile_path
        print(snakefile_path)
    else:
        msg = 'Error: cannot find Snakefile at the following location:\n'
        msg += '{}\n'.format(snakefile_path)
        sys.stderr.write(msg)
        sys.exit(-1)

    dependency_dir = os.path.join(args.dependency_dir, "ML_dependencies")
    workflow = None
    workflow = {"DEPENDENCYDIR": dependency_dir}


    status = snakemake.snakemake(snakefile, config=workflow,
                                targets = ["defense_finder_update"], 
                                use_conda = True,
                                conda_prefix = os.path.join(args.dependency_dir, "ML_dependencies", 'ML_envs'),
                                cores = 1, workdir = cwd)

     # Check if the workflow executed successfully
    if status:
        print("Defensefinder models updated succesfully")
    else:
        print("Update of Defensefinder models failed.")


import pytest
import subprocess
import pandas as pd
import os
import glob
import shutil

def remove_dir_with_contents(path):
    if path in ["/", "C:\\", os.environ["HOME"]]:
        print("Dangerous deletion path, aborting.")
        return
    try:
        logs = glob.glob(f"{path}/logs/*.log")
        for log in logs:
            os.remove(log)
            
        os.rmdir(f"{path}/logs")

        tsvs = glob.glob(f"{path}/*.tsv")
        for tsv in tsvs:
            os.remove(tsv)
        
        bins = glob.glob(f"{path}/*_bins/*.fa")
        for bin in bins:
            os.remove(bin)
        
        bin_dirs = glob.glob(f"{path}/*_bins")
        for bin_dir in bin_dirs:
            os.rmdir(bin_dir)
            
        jsons = glob.glob(f"{path}/*.json")
        for json in jsons:
            os.remove(json)
        
        os.rmdir(path)
            
    except Exception as e:
        print(f"Error removing directory {path}: {e}")


def test_contamination_detection():
    """
    Test contamination detection.
    
    nanomotif detect_contamination \
        --motifs_scored datasets/binnary_testdata/motifs-scored.tsv \
        --bin_motifs datasets/binnary_testdata/bin-motifs.tsv \
        --contig_bins datasets/binnary_testdata/contig_bin.tsv \
        --write_bins --assembly_file datasets/binnary_testdata/assembly_file.fasta \
        -t 1 --out testrun_contamination
    
    """
    outdir = "tests/binnary/testrun_contamination_1"
    bin_contamination_file = f"{outdir}/bin_contamination.tsv"
    
    cmd = [
        "nanomotif", "detect_contamination",
        "--motifs_scored", "datasets/binnary_testdata/motifs-scored.tsv",
        "--bin_motifs", "datasets/binnary_testdata/bin-motifs.tsv",
        "--contig_bins", "datasets/binnary_testdata/contig_bin.tsv",
        "--write_bins",
        "--assembly_file", "datasets/binnary_testdata/assembly_file.fasta",
        "-t", "1",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    # Check that the output directory was created
    assert os.path.isdir(outdir), "Output directory was not created"

    # Check that the TSV file exists
    assert os.path.isfile(bin_contamination_file), "TSV file was not created"

    # Check content
    contamination = pd.read_csv(bin_contamination_file, sep="\t")
    
    # bin	bin_contig_compare	binary_methylation_missmatch_score	non_na_comparisons	contig
    # b3	b3_contig_12	2	4	contig_12
    # b3	b3_contig_13	2	4	contig_13
    # b3	b3_contig_6	3	4	contig_6

    
    assert contamination.shape[0] == 3, "Incorrect number of rows in contamination file"
    assert contamination["bin"].unique() == ["b3"], "Incorrect bin in contamination file"

    # Cleanup after test
    remove_dir_with_contents(outdir)
    
    
def test_contamination_detection_with_file():
    """
    Test contamination detection.
    
    nanomotif detect_contamination \
        --motifs_scored datasets/binnary_testdata/motifs-scored.tsv \
        --bin_motifs datasets/binnary_testdata/bin-motifs.tsv \
        --contig_bins datasets/binnary_testdata/contig_bin.tsv \
        --contamination_file datasets/binnary_testdata/bin_contamination.tsv \
        --write_bins --assembly_file datasets/binnary_testdata/assembly_file.fasta \
        -t 1 --out tests/binnary/testrun_contamination
    
    """
    outdir = "tests/binnary/testrun_contamination_2"
    
    cmd = [
        "nanomotif", "detect_contamination",
        "--motifs_scored", "datasets/binnary_testdata/motifs-scored.tsv",
        "--bin_motifs", "datasets/binnary_testdata/bin-motifs.tsv",
        "--contig_bins", "datasets/binnary_testdata/contig_bin.tsv",
        "--write_bins",
        "--assembly_file", "datasets/binnary_testdata/assembly_file.fasta",
        "--contamination_file", "datasets/binnary_testdata/bin_contamination.tsv",
        "-t", "1",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    # Check that the output directory was created
    assert os.path.isdir(outdir), "Output directory was not created"

    # Check that the detect_contamination_bins/*.fa exists
    bins_in_dir = glob.glob(f"{outdir}/detect_contamination_bins/*.fa")
    assert len(bins_in_dir) == 4, "No bins were written"
    
    
    remove_dir_with_contents(outdir)
    
    
       
def test_include_with_run_detect_contamination():
    """
    Test include contigs with run contamination flag.
    
    nanomotif include_contigs \
        --motifs_scored nanomotif/datasets/binnary_testdata/motifs-scored.tsv \
        --bin_motifs nanomotif/datasets/binnary_testdata/bin-motifs.tsv \
        --contig_bins nanomotif/datasets/binnary_testdata/contig_bin.tsv \
        --run_detect_contamination \
        --write_bins --assembly_file nanomotif/datasets/binnary_testdata/assembly_file.fasta \
        --min_motif_comparisons 2 \
        -t 1 --out tests/binnary/include_contigs
    
    """
    outdir = "tests/binnary/testrun_include_contigs_1"
    
    cmd = [
        "nanomotif", "include_contigs",
        "--motifs_scored", "datasets/binnary_testdata/motifs-scored.tsv",
        "--bin_motifs", "datasets/binnary_testdata/bin-motifs.tsv",
        "--contig_bins", "datasets/binnary_testdata/contig_bin.tsv",
        "--run_detect_contamination",
        "--write_bins",
        "--assembly_file", "datasets/binnary_testdata/assembly_file.fasta",
        "-t", "1",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    # Check that the output directory was created
    assert os.path.isdir(outdir), "Output directory was not created"

    # Check that the include_contigs_bins/*.fa exists
    bins_in_dir = glob.glob(f"{outdir}/include_contigs_bins/*.fa")
    assert len(bins_in_dir) == 4, "No bins were written"
    
    remove_dir_with_contents(outdir)
    
    
           
def test_include_with_contamination_file():
    """
    Test include contigs with contamination file.
    
    nanomotif include_contigs \
        --motifs_scored nanomotif/datasets/binnary_testdata/motifs-scored.tsv \
        --bin_motifs nanomotif/datasets/binnary_testdata/bin-motifs.tsv \
        --contig_bins nanomotif/datasets/binnary_testdata/contig_bin.tsv \
        --contamination_file datasets/binnary_testdata/bin_contamination.tsv \
        --write_bins --assembly_file nanomotif/datasets/binnary_testdata/assembly_file.fasta \
        --min_motif_comparisons 2 \
        -t 1 --out tests/binnary/include_contigs
    
    """
    outdir = "tests/binnary/testrun_include_contigs_2"
    
    cmd = [
        "nanomotif", "include_contigs",
        "--motifs_scored", "datasets/binnary_testdata/motifs-scored.tsv",
        "--bin_motifs", "datasets/binnary_testdata/bin-motifs.tsv",
        "--contig_bins", "datasets/binnary_testdata/contig_bin.tsv",
        "--contamination_file", "datasets/binnary_testdata/bin_contamination.tsv",
        "--write_bins",
        "--assembly_file", "datasets/binnary_testdata/assembly_file.fasta",
        "-t", "1",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    # Check that the output directory was created
    assert os.path.isdir(outdir), "Output directory was not created"
    
    bins_in_dir = glob.glob(f"{outdir}/include_contigs_bins/*.fa")
    assert len(bins_in_dir) == 4, "No bins were written"
    
    remove_dir_with_contents(outdir)
    
    
         
def test_include_no_contamination():
    """
    Test include contigs without running contamination.
    
    nanomotif include_contigs \
        --motifs_scored nanomotif/datasets/binnary_testdata/motifs-scored.tsv \
        --bin_motifs nanomotif/datasets/binnary_testdata/bin-motifs.tsv \
        --contig_bins nanomotif/datasets/binnary_testdata/contig_bin.tsv \
        --write_bins --assembly_file nanomotif/datasets/binnary_testdata/assembly_file.fasta \
        --min_motif_comparisons 2 \
        -t 1 --out tests/binnary/include_contigs
    
    """
    outdir = "tests/binnary/testrun_include_contigs_3"
    
    cmd = [
        "nanomotif", "include_contigs",
        "--motifs_scored", "datasets/binnary_testdata/motifs-scored.tsv",
        "--bin_motifs", "datasets/binnary_testdata/bin-motifs.tsv",
        "--contig_bins", "datasets/binnary_testdata/contig_bin.tsv",
        "--write_bins",
        "--assembly_file", "datasets/binnary_testdata/assembly_file.fasta",
        "-t", "1",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    # Check that the output directory was created
    assert os.path.isdir(outdir), "Output directory was not created"

    # Check that the detect_contamination_bins/*.fa does not exists
    assert os.path.isdir(f"{outdir}/detect_contamination_bins") == False, "Bins were written"
    
    bins_in_dir = glob.glob(f"{outdir}/include_contigs_bins/*.fa")
    assert len(bins_in_dir) == 4, "No bins were written"
    
    remove_dir_with_contents(outdir)
    
    
         
def test_contamination_with_output_scores():
    """
    Test detect_contamination with outputting scores.
    
    nanomotif detect_contamination \
        --motifs_scored nanomotif/datasets/binnary_testdata/motifs-scored.tsv \
        --bin_motifs nanomotif/datasets/binnary_testdata/bin-motifs.tsv \
        --contig_bins nanomotif/datasets/binnary_testdata/contig_bin.tsv \
        --min_motif_comparisons 2 \
        --save_scores \
        -t 1 --out <out>
    
    """
    outdir = "tests/binnary/testrun_savescores"
    
    cmd = [
        "nanomotif", "detect_contamination",
        "--motifs_scored", "datasets/binnary_testdata/motifs-scored.tsv",
        "--bin_motifs", "datasets/binnary_testdata/bin-motifs.tsv",
        "--contig_bins", "datasets/binnary_testdata/contig_bin.tsv",
        "--save_scores",
        "-t", "1",
        "--out", outdir
    ]
    
    result = subprocess.run(cmd)
    
    # Check that the CLI tool executed successfully
    assert result.returncode == 0, "CLI tool did not exit successfully"

    # Check that the output directory was created
    assert os.path.isdir(outdir), "Output directory was not created"
    
    bins_in_dir = glob.glob(f"{outdir}/scores/detect_contamination/*.csv")
    assert len(bins_in_dir) == 13, "No scores were written"
    
    csvs = glob.glob(f"{outdir}/scores/*/*.csv")
    for csv in csvs:
        os.remove(csv)
    
    csvdirs = glob.glob(f"{outdir}/scores/*")
    for csvdir in csvdirs:
        os.rmdir(csvdir)
    
    os.rmdir(f"{outdir}/scores")
    remove_dir_with_contents(outdir)


import polars as pl
import pandas as pd
from nanomotif.binnary import utils as ut



def test_split_bin_contig():
    # Input DataFrame
    input_data = pl.DataFrame({
        "contig_bin": ["bin1", "bin1", "bin_c1"],
        "bin_compare": ["bin1_contig_23", "bin1_contig_40", "bin_c1_contig_23"]
    })

    # Expected output DataFrame
    expected_output = pd.DataFrame({
        "contig_bin": ["bin1", "bin1", "bin_c1"],
        "bin_compare": ["bin1_contig_23", "bin1_contig_40", "bin_c1_contig_23"],
        "contig": ["contig_23", "contig_40", "contig_23"]
    })

    # Run the function
    output = ut.split_bin_contig(input_data)
    output = output.to_pandas()
    print(output)
    
    # Verify the output
    pd.testing.assert_frame_equal(output, expected_output, check_like=True)


import polars as pl
import re

def split_bin_contig(df):
    df = df \
        .with_columns(
            pl.when(pl.col("bin_compare").is_not_null()).then(
                pl.col("bin_compare").str.replace(pl.col("contig_bin").fill_null("").first() + "_", "", literal = True)
            ).over("bin_compare").alias("contig")
        )
        
        
        
        # .with_columns([
        #     pl.col("bin_compare").str.replace(pl.col("bin") + "_", "").alias("contig_bin")
        # ])

    return df


def get_bin(bin_contig):
    return re.search(r"(.+)_contig_\d+", bin_contig).group(1)


def add_compare_df(df, bin_contig):
    df = df.filter(pl.col("bin_compare") == bin_contig)\
            .rename({"bin": "contig_bin"})
    
    return df


import pandas as pd
import polars as pl
import numpy as np
from nanomotif.binnary import data_processing as dp 
from nanomotif.binnary import scoring as sc
from nanomotif.binnary import utils as ut
import logging



def include_contigs(motifs_scored_in_bins, bin_consensus, contamination, args):
    """
    Takes the motifs_scored_in_bins and motifs_of_interest DataFrames and finds unbinned contigs with an exact methylation pattern as the bin.
    
    params:
        motifs_scored_in_bins: pd.DataFrame - DataFrame containing the motifs scored in each bin
        motifs_of_interest: list - List of motifs to be considered from bin_motif_binary
        args: argparse.Namespace - Namespace containing the arguments passed to the script
    
    """
    logger = logging.getLogger(__name__)
    logger.info("Starting include_contigs analysis...")
    
    
    # Remove bins with no methylation in consensus
    bins_w_no_methylation = bin_consensus \
        .group_by("bin") \
        .agg(
            pl.sum("methylation_binary").alias("binary_sum")    
        ) \
        .filter(pl.col("binary_sum") == 0) \
        .select("bin") \
        .unique()["bin"]
    

    bin_consensus = bin_consensus \
        .filter(~pl.col("bin").is_in(bins_w_no_methylation))
    
    # Retain only unbinned contigs or contigs in the contamination file
    contigs_for_comparison = motifs_scored_in_bins \
        .filter(
            (pl.col("bin_contig").str.contains("unbinned")) |
            (pl.col("bin_contig").is_in(contamination["bin_contig_compare"]))
        )
    

    contig_bin_comparison_score, contigs_w_no_methylation = sc.compare_methylation_pattern_multiprocessed(
        motifs_scored_in_bins=contigs_for_comparison,
        bin_consensus=bin_consensus,
        mode="include",
        args=args,
        num_processes=args.threads
    )
    
    contig_bin_comparison_score = ut.split_bin_contig(contig_bin_comparison_score)
    
    logger.info("Assigning contigs to bins...")
    # Filter contigs where motif comparisons are less than args.min_motif_comparisons
    # TODO: looking for 0 comparisons is now redundant. Also remove the column.
    contigs_of_interest = contig_bin_comparison_score \
        .filter(
            pl.col("non_na_comparisons") >= args.min_motif_comparisons,
            (~pl.col("bin_compare").is_in(contigs_w_no_methylation)),  # Remove contigs with no methylation
            (pl.col("binary_methylation_missmatch_score") == 0)        # Retain contigs with no methylation missmatch 
            
        ) \
        .sort("bin","bin_compare")
    
    
    single_contigs = contigs_of_interest \
        .group_by("contig") \
        .agg(
            pl.count("contig").alias("contig_count")
        ) \
        .filter(pl.col("contig_count") == 1)
    
    contigs_of_interest = contigs_of_interest \
        .filter(pl.col("contig").is_in(single_contigs["contig"]))
      

    logger.info("Finished include_contigs analysis.")
    return contigs_of_interest


import nanomotif as nm
from nanomotif.logger import configure_logger
import logging as log
import os
import sys
import shutil
from pathlib import Path
import json
os.environ["POLARS_MAX_THREADS"] = "1"
import polars as pl
from polars import col
import numpy as np
import random
import warnings
from nanomotif._version import __version__


def shared_setup(args, working_dir):
    warnings.filterwarnings("ignore")
    # Check if output directory exists
    if not os.path.exists(args.out):
        os.makedirs(args.out)
    else:
        log.warn(f"Output directory {args.out} already exists")

    # Set up logging
    LOG_DIR = working_dir + "/logs"
    Path(LOG_DIR).mkdir(parents=True, exist_ok=True)
    configure_logger(LOG_DIR + f"/{args.command}.main.log", args.verbose)

    # Log arguments
    log.info(f"nanomotif version: {__version__}")
    log.debug("Arguments:")
    for arg in vars(args):
        log.debug(f"{arg}: {getattr(args, arg)}")
    
    # Save settings
    with open(working_dir + f"/args.{args.command}.json", "w") as f:
        json.dump(vars(args), f, indent=2)
    
    # Set seeds
    nm.seed.set_seed(args.seed)




def find_motifs(args, pileup = None, assembly = None):
    # Run nanomotif
    log.info("Starting nanomotif motif finder")
    if pileup is None:
        log.info("Loading pileup")
        pileup = nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)
    assm_lengths = pl.DataFrame({
        "contig": list(assembly.assembly.keys()),
        "length": [len(contig) for contig in assembly.assembly.values()]
    })
    log.info("Filtering pileup")
    # Filter pileup to contigs with mods, minimum 1 mod per 10kb
    contigs_with_mods = pileup.pileup \
        .groupby(["contig", "mod_type"]) \
        .agg(pl.count()) \
        .join(assm_lengths, on = "contig") \
        .filter(pl.col("count") > pl.col("length")/10000) \
        .get_column("contig").unique().to_list()

    total_contigs = len(assembly.assembly.keys())

    contigs_to_process = [contig for contig in assembly.assembly.keys() if contig in contigs_with_mods]
    pileup = pileup.pileup.filter(pl.col("contig").is_in(contigs_to_process))
    remaining_contigs = pileup.get_column("contig").unique().to_list()

    os.makedirs(args.out + "/temp/", exist_ok=True)
    pileup.write_csv(args.out + "/temp/filtered_pileup.tsv", separator="\t")
    log.info(f"Processing {len(remaining_contigs)} of {total_contigs} contigs")
    log.info("Identifying motifs")
    motifs = nm.evaluate.process_sample_parallel(
            assembly, pileup, 
            threads = args.threads,
            search_frame_size = args.search_frame_size,
            threshold_methylation_confident = args.threshold_methylation_confident,
            threshold_methylation_general = args.threshold_methylation_general,
            threshold_valid_coverage = args.threshold_valid_coverage,
            minimum_kl_divergence = args.minimum_kl_divergence,
            verbose = args.verbose,
            log_dir = args.out + "/logs",
            seed = args.seed
        )
    motifs = pl.DataFrame(motifs)
    if motifs is None:
        log.info("No motifs found")
        return

    log.info("Writing motifs")
    def format_motif_df(df):
        if "model" in df.columns:
            n_mod = [x._alpha for x in df["model"]]
            n_nomod = [x._beta for x in df["model"]]
        motif_iupac = [nm.seq.regex_to_iupac(x) for x in df["motif"]]
        motif_type = [nm.utils.motif_type(x) for x in motif_iupac]

        df_out = df.with_columns([
            pl.Series("motif", motif_iupac),
            pl.Series("motif_type", motif_type)
        ])
        if "model" in df.columns:
            df_out = df_out.with_columns([
                pl.Series("n_mod", n_mod),
                pl.Series("n_nomod", n_nomod)
            ])
        try:
            df_out = df_out.select([
                "contig", "motif", "mod_position", "mod_type", "n_mod", "n_nomod", "motif_type",
                "motif_complement", "mod_position_complement", "n_mod_complement", "n_nomod_complement"
            ])
        except:
            df_out = df_out.select([
                "contig", "motif", "mod_position", "mod_type", "n_mod", "n_nomod", "motif_type",
            ])
        # Subtract prior alpha and beta from n_mod and n_nomod
        if "model" in df.columns:
            df_out = df_out.with_columns([
                pl.col("n_mod") - nm.model.DEFAULT_PRIOR_BETA,
                pl.col("n_nomod") - nm.model.DEFAULT_PRIOR_ALPHA
            ])
        return df_out
    def save_motif_df(df, name):
        df = format_motif_df(df)
        df = df.sort(["contig", "mod_type", "motif"])
        df.write_csv(args.out + "/" + name + ".tsv", separator="\t")
    os.makedirs(args.out + "/precleanup-motifs/", exist_ok=True)
    save_motif_df(motifs, "precleanup-motifs/motifs-raw")

    log.info("Postprocessing motifs")
    motifs_file_name = "precleanup-motifs/motifs"

    log.info(" - Writing motifs")
    motifs = motifs.filter(col("score") > 0.1)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    
    motifs_file_name = motifs_file_name + "-score"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Removing sub motifs")
    motifs = nm.postprocess.remove_sub_motifs(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-sub"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Removing noisy motifs")
    motifs = nm.postprocess.remove_noisy_motifs(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-noise"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Merging motifs")
    motifs = nm.postprocess.merge_motifs_in_df(motifs, pileup, assembly)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-merge"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Joining motif complements")
    motifs = nm.postprocess.join_motif_complements(motifs)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-complement"
    save_motif_df(motifs, motifs_file_name)

    log.info(" - Removing motifs observed less than min count")
    motifs = motifs.filter((pl.col("n_mod") + pl.col("n_nomod")) > args.min_motifs_contig)
    if len(motifs) == 0:
        log.info("No motifs found")
        return
    motifs_file_name = motifs_file_name +   "-mincount"
    save_motif_df(motifs, motifs_file_name)



    save_motif_df(motifs, "motifs")

    log.info("Done finding motifs")
    return format_motif_df(motifs)

def score_motifs(args, pileup = None, assembly = None, motifs = None):
    log.info("Starting nanomotif motif scorer")
    if pileup is None:
        log.info("Loading pileup")
        pileup =  nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)
    if motifs is None:
        log.info("Loading motifs")
        motifs = pl.read_csv(args.motifs, separator="\t")
    
    # Ensure motif are iupac
    motifs.with_columns([
        pl.col("motif").map_elements(lambda x: nm.seq.regex_to_iupac(x)).alias("motif")
    ])

    
    if any([item for item in motifs.columns if "_complement" in item]):
        log.debug("Reformatting motifs to include scoring of complementary motifs")
        motifs_fwd = motifs[[item for item in motifs.columns if "complement" not in item]]
        motifs_rev = motifs[["contig", "motif_type", "mod_type"] + [item for item in motifs.columns if "complement" in item]] \
            .select(pl.all().name.map(lambda col_name: col_name.replace('_complement', ''))) \
            .select(motifs_fwd.columns) \
            .filter(pl.col("motif").is_not_null())
        
        motifs = pl.concat([motifs_fwd, motifs_rev]).unique(["motif", "contig", "mod_type", "mod_position"])
    
    # Convert to regex for motif scoring
    motifs.with_columns([
        pl.col("motif").map_elements(lambda x: nm.seq.iupac_to_regex(x)).alias("motif")
    ])

    log.info("Scoring motifs")
    scored_all = nm.scoremotifs.score_sample_parallel(
        assembly, pileup.pileup, motifs,
        threads = args.threads,
        threshold_methylation_general = args.threshold_methylation_general,
        threshold_valid_coverage = 1,
        verbose = args.verbose,
        log_dir = args.out + "/logs",
        seed = args.seed
        )
    scored_all = nm.postprocess.join_motif_complements(scored_all)

    scored_all = scored_all.unique([
            "motif", "contig", "mod_type", "mod_position"
        ])
    
    scored_all = scored_all.sort(["contig", "mod_type", "motif"])
    scored_all.write_csv(args.out + "/motifs-scored.tsv", separator="\t")
    return scored_all

def bin_consensus(args, pileup = None, assembly = None, motifs = None, motifs_scored = None):
    bins = pl.read_csv(args.bins, separator="\t", has_header=False) \
        .rename({"column_1":"contig", "column_2":"bin"})
    if motifs is None:
        log.info("Loading motifs-scored")
        motifs = pl.read_csv(args.motifs, separator="\t")
    if motifs_scored is None:
        log.info("Loading motifs")
        motifs_scored = pl.read_csv(args.motifs_scored, separator="\t")
    if pileup is None:
        log.info("Loading pileup")
        pileup =  nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    if assembly is None:
        log.info("Loading assembly")
        assembly = nm.load_assembly(args.assembly)

    
    if any([item for item in motifs_scored.columns if "_complement" in item]):
        motifs_fwd = motifs_scored[[item for item in motifs_scored.columns if "complement" not in item]]
        motifs_rev = motifs_scored[["contig", "motif_type", "mod_type"] + [item for item in motifs_scored.columns if "complement" in item]] \
            .select(pl.all().name.map(lambda col_name: col_name.replace('_complement', ''))) \
            .select(motifs_fwd.columns) \
            .filter(pl.col("motif").is_not_null())
        motifs_scored = pl.concat([motifs_fwd, motifs_rev]).unique(["motif", "contig", "mod_type", "mod_position"])
    
    if any([item for item in motifs.columns if "_complement" in item]):
        motifs_fwd = motifs[[item for item in motifs.columns if "complement" not in item]]
        motifs_rev = motifs[["contig", "motif_type", "mod_type"] + [item for item in motifs.columns if "complement" in item]] \
            .select(pl.all().name.map(lambda col_name: col_name.replace('_complement', ''))) \
            .select(motifs_fwd.columns) \
            .filter(pl.col("motif").is_not_null())
        motifs = pl.concat([motifs_fwd, motifs_rev]).unique(["motif", "contig", "mod_type", "mod_position"])


    output = nm.bin_consensus.within_bin_motifs_consensus(pileup.pileup, assembly, motifs, motifs_scored, bins)
    output = nm.bin_consensus.merge_bin_motifs(output, bins, pileup, assembly)

    output = output.rename({"bin":"contig", "n_mod_bin":"n_mod", "n_nomod_bin":"n_nomod"})

    output = nm.postprocess.join_motif_complements(output)

    output = output.rename({"contig":"bin", "n_mod":"n_mod_bin", "n_nomod":"n_nomod_bin"})
    output = output.sort(["bin", "mod_type", "motif"])

    output = output.filter(pl.col("n_mod_bin") + pl.col("n_nomod_bin") > args.min_motifs_bin)
    output.write_csv(args.out + "/bin-motifs.tsv", separator="\t")

def motif_discovery(args):
    # Check if output directory exists
    log.info("Loading required files")
    pileup = nm.load_pileup(args.pileup, threads = args.threads, min_fraction = args.threshold_methylation_general)
    assembly = nm.load_assembly(args.assembly)

    # Find motifs
    log.info("Finding motifs")
    motifs = find_motifs(args, pileup=pileup, assembly=assembly)
    if motifs is None:
        log.info("Stopping workflow")
        return

    # Score all motifs
    log.info("Scoring motifs")
    scored_all = score_motifs(args, pileup=pileup, assembly=assembly, motifs=motifs)

    # Bin consensus
    log.info("Finding bin consensus motifs")
    bin_consensus(args, pileup=pileup, assembly=assembly, motifs=motifs, motifs_scored=scored_all)

    log.info("Done")

def check_install(args):
    
    # Check if output directory exists
    log.info("Loading required files")
    args.out = "nanomotif_install_check"

    pileup = nm.datasets.geobacillus_plasmids_pileup()
    assembly = nm.datasets.geobacillus_plasmids_assembly()

    # Find motifs
    log.info("Finding motifs")
    motifs = find_motifs(args, pileup=pileup, assembly=assembly)

    # Score all motifs
    log.info("Scoring motifs")
    scored_all = score_motifs(args, pileup=pileup, assembly=assembly, motifs=motifs)

    # Bin consensus
    log.info("Finding bin consensus motifs")
    args.bins = nm.datasets.geobacillus_plasmids_bin_path()
    bin_consensus(args, pileup=pileup, assembly=assembly, motifs=motifs, motifs_scored=scored_all)
    
    log.info("Done")
    shutil.rmtree(args.out)


####################################################################################################
# Binnary - contamination and inclusion
from nanomotif.binnary import data_processing, detect_contamination, include_contigs
from nanomotif.binnary.logging import set_logger_config



def binnary(args):
    """
    binnary entry point for the DNA Methylation Pattern Analysis tool.
    Orchestrates the workflow of the tool based on the provided arguments.
    """
    
    # Conditional check for --write_bins and --assembly_file
    if args.write_bins and not args.assembly_file:
        print("Error: --assembly_file must be specified when --write_bins is used.")
        sys.exit(1)
    elif not args.write_bins and args.assembly_file:
        print("Error: --assembly_file can only be used when --write_bins is specified.")
        sys.exit(1)
    
    print("Starting Binnary ", args.command, " analysis...")

    
    # Settting up the logger
    # set_logger_config(args)
    # logger = log.getLogger(__name__)
    log.info("Starting Binnary analysis...")
    
    # Set number of threads for polars
    os.environ['POLARS_MAX_THREADS'] = str(args.threads)
    
    POLAR_THREADS = pl.threadpool_size()
    log.info(f"Polars is using {POLAR_THREADS} threads.")
    
    # Step 1: Load and preprocess data
    # These functions would be defined in your data_processing module
    (
        motifs_scored,
        bin_motifs,
        contig_bins,
    ) = data_processing.load_data(args)

    

    # Step 2: create motifs_scored_in_bins and bin_motif_binary
    bin_motif_binary = data_processing.prepare_bin_consensus(bin_motifs, args)
    
    motifs_in_bin_consensus = bin_motif_binary.select("motif_mod").unique()["motif_mod"]
    
    motifs_scored_in_bins = data_processing.prepare_motifs_scored_in_bins(
        motifs_scored,
        motifs_in_bin_consensus,
        contig_bins
    )
    
    
    # Create the bin_consensus dataframe for scoring
    log.info("Creating bin_consensus dataframe for scoring...")
    bin_motifs_from_motifs_scored_in_bins = data_processing.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )

    # Setting up the contamination analysis
    if (args.command == "detect_contamination" and not args.contamination_file) or (args.command == "include_contigs" and args.run_detect_contamination):
        contamination = detect_contamination.detect_contamination(
            motifs_scored_in_bins, bin_motifs_from_motifs_scored_in_bins, args
        )
        data_processing.generate_output(contamination.to_pandas(), args.out, "bin_contamination.tsv")
    elif args.contamination_file:
        log.info("Loading contamination file...")
        contamination = data_processing.load_contamination_file(args.contamination_file)
        
    if args.command == "detect_contamination":
        # Create a decontaminated contig_bin file
        new_contig_bins = data_processing.create_contig_bin_file(
            contig_bins=contig_bins.to_pandas(), 
            contamination=contamination.to_pandas(),
            include=None
        )
        data_processing.generate_output(new_contig_bins, args.out, "decontaminated_contig_bin.tsv")

    if args.command == "include_contigs":
        # # User provided contamination file
        # if args.contamination_file:
        #     log.info("Loading contamination file...")
        #     contamination = data_processing.load_contamination_file(args.contamination_file)
        
        ## If run_detect_contamination is false and contamination file is not provided, then set contamination to None
        if not args.run_detect_contamination and not args.contamination_file:
            contamination = pl.DataFrame(
                {
                    "bin": [],
                    "bin_contig_compare": [],
                    "binary_methylation_missmatch_score": [],
                    "non_na_comparisons": [],
                    "contig": []
                }
            )
        
        
        
        # Run the include_contigs analysis    
        include_contigs_df = include_contigs.include_contigs(
            motifs_scored_in_bins, bin_motifs_from_motifs_scored_in_bins, contamination, args
        )
        
        # Save the include_contigs_df results
        data_processing.generate_output(include_contigs_df.to_pandas(), args.out, "include_contigs.tsv")
        
        # Create a new contig_bin file
        
        
        new_contig_bins = data_processing.create_contig_bin_file(
            contig_bins=contig_bins.to_pandas(), 
            contamination= contamination.to_pandas(),
            include=include_contigs_df.to_pandas()
        )
        data_processing.generate_output(new_contig_bins, args.out, "new_contig_bin.tsv")
        
    if args.write_bins:
        log.info("Write bins flag is set. Writing bins to file...")
        log.info("Loading assembly file...")
        assembly = data_processing.read_fasta(args.assembly_file)
        
        bin_dir = os.path.join(args.out, args.command + "_bins")
        
        data_processing.write_bins_from_contigs(new_contig_bins, assembly, bin_dir)
    
    log.info(f"Analysis Completed. Results are saved to: {args.out}")
    print("Analysis Completed. Results are saved to:", args.out)


from nanomotif.mtase_linker.dependencies import snakemake_create_environments, get_models, defensefinder_update
from nanomotif.mtase_linker.command import run_MTase_linker

def mtase_linker(args):
    if args.mtase_linker_command == "install":
        snakemake_create_environments(args)
        get_models(args)
        defensefinder_update(args)
    elif args.mtase_linker_command == "run":
        run_MTase_linker(args)
    else:
        print(f"Unknown MTase-linker command: {args.mtase_linker_command}")
       



def main():
    # Parse arguments
    parser = nm.argparser.create_parser()
    args = parser.parse_args()
    
    if args.command in ["detect_contamination", "include_contigs", "MTase-linker"]:
        args.verbose = False
        args.seed = 1
    
    if args.command == "find_motifs":
        shared_setup(args, args.out)
        find_motifs(args)
    elif args.command == "score_motifs":
        shared_setup(args, args.out)
        score_motifs(args)
    elif args.command == "bin_consensus":
        shared_setup(args, args.out)
        bin_consensus(args)
    elif args.command == "motif_discovery":
        shared_setup(args, args.out)
        motif_discovery(args)

    elif args.command in ["detect_contamination", "include_contigs"]:
        shared_setup(args, args.out)
        binnary(args)

    elif args.command == "MTase-linker":
        mtase_linker(args)

    elif args.command == "check_installation":
        args.out = "nanomotif_install_check"
        shared_setup(args, args.out)
        check_install(args)

    else:
        parser.print_help()
        exit()
if __name__ == "__main__":
    main()



import pytest
from nanomotif.argparser import create_parser


def test_argparser_detect_contamination_with_specific_data(mocker):
    # Mock sys.argv with the specific command line you provided
    mocker.patch(
        "sys.argv",
        [
            "nanomotif",  # Assuming 'binnary.py' is a typo and should be 'binary.py' or similar
            "detect_contamination",
            "--motifs_scored",
            "data/motifs-scored.tsv",
            "--bin_motifs",
            "data/bin-motifs.tsv",
            "--contig_bins",
            "data/bins.tsv",
            "--out",
            "output",
        ],
    )

    # Parse the arguments
    parser = create_parser()
    args = parser.parse_args()

    # Assertions to verify that all arguments are correctly parsed
    assert args.command == "detect_contamination"
    assert args.motifs_scored == "data/motifs-scored.tsv"
    assert args.bin_motifs == "data/bin-motifs.tsv"
    assert args.contig_bins == "data/bins.tsv"
    assert args.out == "output"



import pytest
from nanomotif.binnary import detect_contamination
from nanomotif.binnary import data_processing as dp
from .conftest import MockArgs
import os


def test_detect_contamination(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    GIVEN loaded_data
    WHEN detect_contamination is called
    THEN assert that the output contains only contig 3 and the expected columns
    """
    args = MockArgs()
    
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    
    bin_motifs_from_motifs_scored_in_bins = dp.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )
    
    contaminated_contigs = detect_contamination.detect_contamination(
        motifs_scored_in_bins,
        bin_motifs_from_motifs_scored_in_bins,
        args
    )
    print(contaminated_contigs)
    contaminated_contigs = contaminated_contigs.to_pandas()
    
    assert contaminated_contigs is not None
    assert contaminated_contigs["bin"].unique().tolist() == ["b3"]
    assert sorted(contaminated_contigs["contig"].unique().tolist()) == ["contig_12", "contig_13", "contig_6"]
    assert contaminated_contigs[contaminated_contigs["contig"] == "contig_6"]["binary_methylation_missmatch_score"].values[0] == 3.0


def test_load_contamination_file(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    Test that the contamination file is loaded correctly.
    """
    # setup
    args = MockArgs()
    
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    
    bin_motifs_from_motifs_scored_in_bins = dp.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )
    
    contaminated_contigs = detect_contamination.detect_contamination(
        motifs_scored_in_bins,
        bin_motifs_from_motifs_scored_in_bins,
        args
    )
    
    contaminated_contigs = contaminated_contigs.to_pandas().to_csv("./tests/contaminated_contigs.tsv", index=False,sep="\t")
    
    # test
    
    contamination_file = dp.load_contamination_file("./tests/contaminated_contigs.tsv")
    required_columns = ["bin", "bin_contig_compare", "binary_methylation_missmatch_score", "non_na_comparisons", "contig"]
    
    contamination_file = contamination_file.to_pandas()
    
    assert not contamination_file.empty, "The contamination file should not be empty."
    assert list(contamination_file.columns) == required_columns, "The contamination file columns are incorrect."
    
    os.remove("./tests/contaminated_contigs.tsv")
    
    


import pandas as pd
import numpy as np
import itertools
from scipy.spatial.distance import euclidean

from .data_processing import prepare_motifs_scored_in_bins


def perform_analysis(motifs_scored, bin_motifs, contig_bins, assembly_stats, assembly_file, args):
    """
    Performs the core analysis of the tool. This function is called from the main entry point.
    """
    # Find relevant motifs in bins and contigs
    motifs_scored_in_bins = prepare_motifs_scored_in_bins(motifs_scored, bin_motifs, contig_bins, assembly_stats)
    
    # Calculate belonging score
    belonging_score = calculate_belonging_score(motifs_scored_in_bins, args)
    
    # Determine contamination
    belonging_score = determine_contamination(belonging_score)
    
    # Calculate kmer distance for contigs that can belong to multuple bins
    contamination_kmer_distance_dict = calculate_contamination_kmer_distance(belonging_score, assembly_file, contig_bins, args)
    # Initialize a list to hold your formatted data
    formatted_data = []

    # Iterate over the dictionary to format the string
    for contig, bins in contamination_kmer_distance_dict.items():
        # Format the bin distances into a single string
        matching_bins = '|'.join([f"{bin}:{distance}" for bin, distance in bins.items()])
        # Append the contig and the formatted string to your data list
        formatted_data.append([contig, matching_bins])

    # Create a DataFrame from the formatted data
    contamination_kmer_distance_df = pd.DataFrame(formatted_data, columns=['contig', 'matching_bins'])

    
    # Dataframe for output
    output_df = belonging_score[["bin", "contig", "bin_id", "contamination_group"]]\
        .rename(columns={"bin": "matching_bins", "bin_id": "current_bin"})
    
    ## Merge contamination_df with contamination_kmer_distance
    contamination_matching_multiple_bins = output_df[
        (output_df["contamination_group"] == "multiple_contamination") & (output_df["current_bin"] != "unbinned")
    ].drop(columns=["matching_bins"])\
    .drop_duplicates(subset='contig')\
    .merge(contamination_kmer_distance_df, on="contig", how="left")\
    .dropna()
    
    ## 
    unbinned_matching_multiple_bins = output_df[
        (output_df["contamination_group"] == "multiple_contamination") & (output_df["current_bin"] == "unbinned")
    ]
    unbinned_matching_multiple_bins = unbinned_matching_multiple_bins.groupby('contig')['matching_bins'].apply(lambda x: '|'.join(x)).reset_index()
    unbinned_matching_multiple_bins.rename(columns={'matching_bins': 'matching_bins'}, inplace=True)
    unbinned_matching_multiple_bins["current_bin"] = "unbinned"
    
    correct_assignment_df = output_df[output_df["contamination_group"] == "correct"]
    
    ## Concat all dataframes
    output_final = pd.concat([correct_assignment_df, contamination_matching_multiple_bins, unbinned_matching_multiple_bins])\
        .drop(columns=["contamination_group"])
    
    return output_final





def calculate_belonging_score(motifs_scored_in_bins, args):
    """
    Finds the closest match between contig and bin based on methylation pattern.
    """
    # Step 1: Group by bin and motif_mod and calculate mean methylation
    bin_motif_binary = motifs_scored_in_bins[motifs_scored_in_bins['bin'] != "unbinned"].groupby(['bin', 'motif_mod'])['mean'].mean().reset_index(name='mean_methylation')
    
    # Step 2: Convert mean methylation values to binary
    bin_motif_binary['methylation_binary'] = (bin_motif_binary['mean_methylation'] >= args.mean_methylation_cutoff).astype(int)

    
    # Create contig_motif_binary
    ## Filter motifs that are not observed in bins
    contig_motif_binary = motifs_scored_in_bins[motifs_scored_in_bins["motif_mod"].isin(bin_motif_binary["motif_mod"])]
    
    ## Filter motifs that are not observed more than n_motif_cutoff times
    contig_motif_binary = contig_motif_binary[contig_motif_binary["n_motifs"] >= args.n_motif_cutoff]
    
    ## Convert mean methylation values to binary
    contig_motif_binary["methylation_binary"] = (contig_motif_binary["mean"] >= args.mean_methylation_cutoff).astype(int)
    
    ## Rename bin_contig to bin
    contig_motif_binary = contig_motif_binary[["bin_contig", "motif_mod", "methylation_binary"]]
    contig_motif_binary.rename(columns={"bin_contig": "bin"}, inplace=True)
    
    # Pivot the DataFrame
    contig_motif_binary_pivoted = contig_motif_binary.pivot_table(
        index='bin', columns='motif_mod', values='methylation_binary', fill_value=None
    )
    # Unpivot the DataFrame back to long format
    contig_motif_binary = contig_motif_binary_pivoted.reset_index().melt(
        id_vars=['bin'], var_name='motif_mod', value_name='methylation_binary'
    )
    
    # Check if there is a bin with no methylated motifs
    bin_with_no_methylations_exists = bin_motif_binary.groupby("bin")["methylation_binary"].sum().min() == 0
    
    if bin_with_no_methylations_exists == False:
        print("No bin with no methylated motifs exists. Removing contigs with no methylated motifs...")
        # Remove contigs with no methylated motifs
        contig_motif_binary = contig_motif_binary[contig_motif_binary.groupby("bin")["methylation_binary"].transform("sum") > 0]

    
    # Combine bin_motif_binary and contig_motif_binary
    contig_motif_binary = contig_motif_binary.rename(columns={
        'bin': 'bin_compare',
        'methylation_binary': 'methylation_binary_compare'
    })
    
    motif_binary_compare = pd.merge(bin_motif_binary, contig_motif_binary, on='motif_mod')
    
    # Define the conditions
    conditions = [
        (motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'] == 1),
        (motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'] == 0),
        (motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'] == 1),
        (motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'] == 0),
        (motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'].isna()),
        (motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'].isna())
    ]

    # Define the corresponding choices for each condition
    choices = [1, -1, -1, 0, 0, 0]

    # Use numpy.select to apply these conditions and choices
    motif_binary_compare['motif_comparison_score'] = np.select(conditions, choices, default=np.nan)
    
    
    # Calculate belonging score
    ## Calculate the sum of motif_comparison_score for each combination of bin and bin_compare
    belonging_score = motif_binary_compare.groupby(["bin", "bin_compare"])["motif_comparison_score"].sum().reset_index(name="belonging_score")
    
    # Find the max belonging_score for each bin_compare
    max_scores = belonging_score.groupby("bin_compare")["belonging_score"].max().reset_index(name="max_belonging_score")

    # Merge the max_scores back with the original belonging_score DataFrame to get the corresponding bin values
    # This merge operation ensures that we only retain the rows with the max belonging_score for each bin_compare
    belonging_score_with_max = pd.merge(belonging_score, max_scores, how='inner', left_on=["bin_compare", "belonging_score"], right_on=["bin_compare", "max_belonging_score"])

    # Count the number of best matches for each bin_compare
    belonging_score_with_max["belonging_bins"] = belonging_score_with_max.groupby("bin_compare")["bin_compare"].transform("count")
    
    
    # Drop the max_belonging_score column as it's redundant now
    belonging_score_final = belonging_score_with_max.drop(columns=["max_belonging_score"])

    separated_columns = belonging_score_final['bin_compare'].str.split('_', expand=True)
    belonging_score_final[['bin_id', 'contig', 'contig_number', 'length']] = separated_columns

    # Step 2: Modify 'contig' column
    belonging_score_final['contig'] = 'contig_' + belonging_score_final['contig_number']

    # Step 3: Convert 'length' column to numeric
    belonging_score_final['length'] = pd.to_numeric(belonging_score_final['length'])

    # Step 4: Remove 'contig_number' column
    belonging_score_final.drop(columns=['contig_number'], inplace=True)
    
    
    return belonging_score_final


def determine_contamination(belonging_score):
    """
    Determines whether a contig is contaminated based on belonging_score.
    """
    # Define the conditions
    conditions = [
        (belonging_score["bin"] == belonging_score["bin_id"]) & (belonging_score["belonging_bins"] == 1),
        (belonging_score["bin"] != belonging_score["bin_id"]) & (belonging_score["belonging_bins"] == 1),
        (belonging_score["belonging_bins"] > 1)
    ]

    # Define the corresponding choices for each condition
    choices = [
        "correct", # pattern between bin and contig matches and there is only one best match 
        "single_contamination", # contig pattern does not match bin pattern and there is only one best match
        "multiple_contamination" # contig pattern does not match bin pattern and there are multiple best matches
    ]

    belonging_score["contamination_group"] = np.select(conditions, choices, default=np.nan)
    
    return belonging_score
    


def calculate_contamination_kmer_distance(belonging_score, assembly_file, contig_bins, args):
    """
    Calculates the kmer distance between contigs and bins that they belong to.
    """
    # Intialize a distance dictionary
    distances = {}
    
    # Find contigs that matches multiple bins based on methylation pattern
    contigs_in_bins_matching_multiple_bins = belonging_score[(belonging_score["contamination_group"] == "multiple_contamination") & (belonging_score["bin_id"] != "unbinned")]["contig"].unique()
    
    for contig in contigs_in_bins_matching_multiple_bins:
        # Initialize a dictionary for this contig's distances
        contig_distances = {}
        
        # Find matching bins
        matching_bins = belonging_score[belonging_score["contig"] == contig]["bin"].unique()

        # Generate kmer_df
        kmers = generate_kmers(args.kmer_window_size)
        kmer_df = pd.DataFrame(index=kmers)
        
        # count kmer for each matching bin
        for bin in matching_bins:
            # Initialize a Series to hold summed k-mer counts for the bin
            bin_kmer_counts = pd.Series(0, index=kmers)
            
            # Get contigs in bin
            contigs_in_bin = contig_bins[contig_bins["bin"] == bin]["contig"].unique()
            
            # remove contig of interest from contigs_in_bin
            contigs_in_bin = contigs_in_bin[contigs_in_bin != contig]
            
            # calculate kmer frequency for each contig in bin from the assembly_file dictionary
            for contig_in_bin in contigs_in_bin:
                # Get contig sequence
                contig_sequence = assembly_file.get(contig_in_bin, "")
                # Count kmer for contig
                contig_kmer_counts = count_kmers(contig_sequence, args.kmer_window_size)
                # Add contig_kmer_counts to bin_kmer_counts
                bin_kmer_counts += pd.Series(contig_kmer_counts).reindex(bin_kmer_counts.index, fill_value=0)
            
            # Add summed k-mer counts for the bin to the DataFrame
            kmer_df[bin] = bin_kmer_counts
            
        # Count k-mer for contig of interest
        contig_sequence = assembly_file.get(contig, "")
        contig_kmer_counts = count_kmers(contig_sequence, args.kmer_window_size)
        kmer_df[contig] = pd.Series(contig_kmer_counts).reindex(kmer_df.index, fill_value=0)
        
        # Convert to kmer frequency
        total_kmer_counts = kmer_df.sum(axis=0)
        kmer_df = kmer_df.div(total_kmer_counts, axis=1)
        
        # Calculate distance between contig column and each bin column
        # Iterate over columns in the DataFrame
        for col in kmer_df.columns:
            if col != contig:  # Skip the contig column itself
                # Calculate the distance between the contig column and the current bin column
                distance = euclidean(kmer_df[contig], kmer_df[col])
                # Store the distance in the dictionary
                contig_distances[col] = distance
        
        # After processing all bins, store the contig's distances in the main dictionary
        distances[contig] = contig_distances
        
    return distances



def generate_kmers(k, alphabet='ATCG'):
    """Generate all possible k-mers from the given alphabet."""
    return [''.join(p) for p in itertools.product(alphabet, repeat=k)]

def count_kmers(sequence, k):
    """Count k-mers in a given sequence."""
    kmer_counts = {kmer: 0 for kmer in generate_kmers(k)}
    for i in range(len(sequence) - k + 1):
        kmer = str(sequence[i:i+k])
        if kmer in kmer_counts:  # This check ensures we only count valid k-mers
            kmer_counts[kmer] += 1
    return kmer_counts




from nanomotif.binnary import data_processing as dp
from nanomotif.binnary import utils as ut
from nanomotif.parallel import update_progress_bar
import os
import polars as pl
import multiprocessing
from multiprocessing import get_context
import logging
from logging.handlers import QueueHandler, QueueListener


def define_mean_methylation_thresholds(motif_binary_compare):
    """
    Define mean methylation thresholds for bin and contig motifs
    """
    # Calculate the mean methylation value for each motif in each bin
    motif_binary_compare = motif_binary_compare.with_columns([
        pl.when(pl.col("methylation_binary") == 1)
        .then(
            pl.when(pl.col("mean_methylation") - 4 * pl.col("std_methylation_bin") > 0.1)
            .then(pl.col("mean_methylation") - 4 * pl.col("std_methylation_bin"))
            .otherwise(0.1)
        )
        .otherwise(pl.lit(None))
        .alias("methylation_mean_threshold")
    ])

    # Calculate the binary methylation value for each motif in each bin where the bin consensus is 1
    motif_binary_compare = motif_binary_compare.with_columns([
        pl.when((pl.col("methylation_binary") == 1) & 
                ((pl.col("mean") >= pl.col("methylation_mean_threshold")) | 
                (pl.col("mean") > 0.4)))
        .then(1)
        .when(pl.col("methylation_binary") == 1)
        .then(0)
        .otherwise(pl.lit(None))
        .alias("methylation_binary_compare")
    ])

    # Calculate score for bin consensus is 0
    motif_binary_compare = motif_binary_compare.with_columns([
        pl.when(pl.col("methylation_binary") == 0)
        .then(0.25)
        .otherwise(pl.col("methylation_mean_threshold"))
        .alias("methylation_mean_threshold"),

        pl.when(pl.col("methylation_binary") == 0)
        .then((pl.col("mean") >= 0.25).cast(pl.Int32))
        .otherwise(pl.col("methylation_binary_compare"))
        .alias("methylation_binary_compare")
    ])
    
    return motif_binary_compare



def compare_methylation_pattern(motif_binary_compare):
    """
    Compares the methylation pattern between bin and contig motifs using Polars and calculates the motif_comparison_score.
    """
    motif_comparison_score = (
        pl.when((motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'] == 1))
        .then(0)
        .when((motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'] == 0))
        .then(1)
        .when((motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'] == 1))
        .then(1)
        .when((motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'] == 0))
        .then(0)
        .when((motif_binary_compare['methylation_binary'] == 1) & (motif_binary_compare['methylation_binary_compare'].is_null()))
        .then(0)
        .when((motif_binary_compare['methylation_binary'] == 0) & (motif_binary_compare['methylation_binary_compare'].is_null()))
        .then(0)
        .otherwise(pl.lit(None))
    )

    # Add the 'motif_comparison_score' column to the DataFrame
    motif_binary_compare = motif_binary_compare.with_columns(motif_comparison_score.alias("motif_comparison_score"))
    
    # Group by bin and bin_compare and calculate the sum of the motif_comparison_score and the count of non-NA values
    contig_bin_comparison_score = motif_binary_compare \
        .group_by(["bin", "bin_compare", "contig_bin"]) \
        .agg([
            pl.sum("motif_comparison_score").alias("binary_methylation_missmatch_score"),
            pl.count("motif_comparison_score").alias("non_na_comparisons")
        ])
        
    return contig_bin_comparison_score


def process_bin_contig(
    task, 
    bin_motifs_from_motifs_scored_in_bins, 
    motifs_scored_in_contigs, 
    mode, 
    args,
    counter,
    lock
):
    """
    This function processes a single contig and compares the methylation pattern between the contig and the bin.
    Depending on the mode, the function will either look for contamination or include contigs.
    
    Parameters:
    - bin_contig: str
        The contig to process
    - bin_motifs_from_motifs_scored_in_bins: polars.DataFrame
        The motifs scored in the bin
    - motifs_scored_in_contigs: polars.DataFrame
        The motifs scored in the contigs
    - mode: str
        The mode to run the comparison in. Either "contamination" or "include"
    - args: argparse.Namespace
        The arguments passed to the script
    
    Returns:
    - contig_bin_comparison_score: polars.DataFrame
        The comparison score for the contig
    """
    
    bin, bin_contig = task
    
    try:
        if mode == "contamination":
            motif_binary_compare = bin_motifs_from_motifs_scored_in_bins \
                .filter(pl.col("bin") == bin) \
                .join(
                    ut.add_compare_df(motifs_scored_in_contigs, bin_contig),# motifs_scored_in_contigs.filter(pl.col("bin_compare") == bin_contig),
                    on="motif_mod"
                )
        if mode == "include":
            motif_binary_compare = bin_motifs_from_motifs_scored_in_bins \
                .filter(pl.col("bin") != bin) \
                .join(
                    ut.add_compare_df(motifs_scored_in_contigs, bin_contig),# motifs_scored_in_contigs.filter(pl.col("bin_compare") == bin_contig),
                    on="motif_mod"
                )
        
        # Define methylation thresholds
        motif_binary_compare = define_mean_methylation_thresholds(motif_binary_compare)

        if args.save_scores:
            path = os.path.join(args.out, "scores", args.command, "binary_compare" + bin_contig + ".csv")
            motif_binary_compare.write_csv(path)

        # Calculate the comparison score regardless of methylation presence
        contig_bin_comparison_score = compare_methylation_pattern(motif_binary_compare)
        
        # Check if the contig has no methylation and note it, but do not exclude it from further processing
        contigHasNMethylation = motif_binary_compare.filter(pl.col("methylation_binary_compare") == 1).height
        
        
        if mode == "include":
            contig_bin_comparison_score = contig_bin_comparison_score \
                .filter(
                    pl.col("binary_methylation_missmatch_score") == 0
                )
        if mode == "contamination":
            contig_bin_comparison_score = contig_bin_comparison_score \
                .filter(
                    pl.col("binary_methylation_missmatch_score") > 0
                )
                
        with lock:
            counter.value += 1
            
        if contigHasNMethylation == 0:
            return contig_bin_comparison_score, bin_contig
        
        return contig_bin_comparison_score, None
    except Exception as e:
        with lock:
            counter.value += 1
            
        error_log_path = os.path.join(args.out, "logs", f"{args.command}_{bin_contig}_error.err")
        
        with open(error_log_path, 'w') as error_file:
            error_file.write(f"Error processing {bin_contig}: {str(e)}\n")
        return None, None

def compare_methylation_pattern_multiprocessed(motifs_scored_in_bins, bin_consensus, mode, args, num_processes=1):
    logger = logging.getLogger(__name__)
    logger.info("Starting comparison of methylation patterns")
    
    motifs_scored_in_contigs = motifs_scored_in_bins \
        .filter(pl.col("n_motifs") >= args.n_motif_contig_cutoff)
    
    unique_tasks = motifs_scored_in_contigs.select(["bin", "bin_contig"]).unique()
    
    tasks = [(row['bin'], row['bin_contig']) for row in unique_tasks.iter_rows(named = True)]
    
    motifs_scored_in_contigs = motifs_scored_in_contigs \
        .select(["bin", "bin_contig", "motif_mod", "mean"]) \
        .rename({"bin_contig": "bin_compare"})
    
    # Create a progress manager
    manager = multiprocessing.Manager()
    counter = manager.Value('i', 0)
    lock = manager.Lock()

    # Create a pool of workers
    pool = get_context("spawn").Pool(processes=num_processes)

    # Create a process for the progress bar
    progress_bar_process = multiprocessing.Process(target=update_progress_bar, args=(counter, len(tasks), True))
    progress_bar_process.start()
    
    
    
    if args.save_scores:
        dir = os.path.join(args.out, "scores", args.command)
        if not os.path.exists(dir):
            os.makedirs(dir)

    # Put them workers to work
    results = pool.starmap(process_bin_contig, [(
        task, 
        bin_consensus, 
        motifs_scored_in_contigs, 
        mode, 
        args,
        counter,
        lock
        ) for task in tasks])
    results = [result for result in results if result is not None] #TODO: Check if this is necessary

    # Close the pool
    pool.close()
    pool.join()

    # Close the progress bar
    progress_bar_process.join()
    
    comparison_score = pl.DataFrame()
    contigs_w_no_methylation = []
    
    for result, no_methylation in results:
        if result is not None:
            comparison_score = pl.concat([comparison_score, result])
        if no_methylation is not None:
            contigs_w_no_methylation.append(no_methylation)
        
    return comparison_score, contigs_w_no_methylation




import pytest
from nanomotif.binnary import include_contigs as ic
from nanomotif.binnary import detect_contamination as dc
from nanomotif.binnary import data_processing as dp
from .conftest import MockArgs

def test_include_contigs(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    GIVEN loaded_data
    WHEN include_contigs is called
    THEN assert that the output contains only contig 3 and the expected columns
    """
    args = MockArgs()
    
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    
    
    bin_motifs_from_motifs_scored_in_bins = dp.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )
    
    contamination = dc.detect_contamination(motifs_scored_in_bins, bin_motifs_from_motifs_scored_in_bins, args)
    
    include_contigs_df = ic.include_contigs(motifs_scored_in_bins, bin_motifs_from_motifs_scored_in_bins, contamination, args)
    
    include_contigs_df = include_contigs_df.to_pandas()
    
    assert include_contigs_df is not None
    assert include_contigs_df.shape[0] == 3
    assert include_contigs_df[include_contigs_df["bin"] == "b3"]["contig"].iloc[0] == "contig_15"
    
    
    
def test_include_with_too_high_min_motif_comparisons(loaded_data, motifs_scored_in_bins_and_bin_motifs):
    """
    GIVEN loaded_data
    WHEN include_contigs is called with a min_motif_comparisons of 10
    THEN assert that the output contains only contig 3 and the expected columns
    """
    args = MockArgs()
    args.min_motif_comparisons = 10
    
    
    
    motifs_scored_in_bins = motifs_scored_in_bins_and_bin_motifs["motifs_scored_in_bins"]
    bin_motifs_from_motifs_scored_in_bins = dp.construct_bin_consensus_from_motifs_scored_in_bins(
        motifs_scored_in_bins,
        args
    )
    
    contamination = dc.detect_contamination(motifs_scored_in_bins, bin_motifs_from_motifs_scored_in_bins, args)
    
    include = ic.include_contigs(motifs_scored_in_bins, bin_motifs_from_motifs_scored_in_bins, contamination, args)
    
    # Assert df is empty
    assert include.shape[0] == 0